#include "bitmap.h"
#include <olectl.h>

HBITMAP CreateBitmap(IImage* pimgSrc, SIZE& sz)
{
	PixelFormat pf;
	pf.dwSize = sizeof(PixelFormat);
	DWORD dwCX, dwCY;
	pimgSrc->GetSize(&dwCX, &dwCY);
	pimgSrc->GetFormat(&pf);
	BITMAPINFO bi;
	RGBQUAD* prgbqPixels;
	void* ppixSrc;
	HDC hdc = ::GetDC(::GetDesktopWindow());
	bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bi.bmiHeader.biWidth = dwCX;
	bi.bmiHeader.biHeight = -(int)dwCY;
	bi.bmiHeader.biPlanes = 1;
	bi.bmiHeader.biBitCount = 32;
	bi.bmiHeader.biCompression = BI_RGB;
	bi.bmiHeader.biSizeImage = 0;
	bi.bmiHeader.biXPelsPerMeter = 640;
	bi.bmiHeader.biYPelsPerMeter = 640;
	bi.bmiHeader.biClrUsed = 0;
	bi.bmiHeader.biClrImportant = 0;
	HBITMAP hbm = ::CreateDIBSection(hdc, &bi, DIB_RGB_COLORS, (void**)&prgbqPixels, NULL, 0);
	if(!hbm)
		return NULL;
	::ReleaseDC(::GetDesktopWindow(), hdc);
	pimgSrc->LockImage(&ppixSrc);
	RGBQUAD* pDest = prgbqPixels + (dwCX * dwCY);
	DWORD i, j;
	switch(pf.dwTypeFlags)
	{
	case PF_FLOAT | PF_LONG:
		{
			double * pdSrc = (double*)ppixSrc;
			if(pf.bChannels == 1)
			{
				for(j=0; j < dwCY; j++)
				{
					pDest -= dwCX;
					for(i=0; i < dwCX; i++, pdSrc++, pDest++)
					{
						pDest->rgbRed = (BYTE)(255.0 * pdSrc[0]);
						pDest->rgbGreen = (BYTE)(255.0 * pdSrc[0]);
						pDest->rgbBlue = (BYTE)(255.0 * pdSrc[0]);
						pDest->rgbReserved  = 0x00;
					}
					pDest -= dwCX;
				}
			}
			else
			{
				for(j=0; j < dwCY; j++)
				{
					pDest -= dwCX;
					for(i=0; i < dwCX; i++, pdSrc+=4, pDest++)
					{
						pDest->rgbRed = (BYTE)(255.0 * pdSrc[0]);
						pDest->rgbGreen = (BYTE)(255.0 * pdSrc[1]);
						pDest->rgbBlue = (BYTE)(255.0 * pdSrc[2]);
						pDest->rgbReserved  = 0x00;
					}
					pDest -= dwCX;
				}
			}
		} break;
	case PF_INT | PF_BYTE :
		{
			BYTE * pbSrc = (BYTE*)ppixSrc;
			if(pf.bChannels == 1)
			{
				for(j=0; j < dwCY; j++)
				{
					pDest -= dwCX;
					for(i=0; i < dwCX; i++, pbSrc++, pDest++)
					{
						pDest->rgbRed = pbSrc[0];
						pDest->rgbGreen = pbSrc[0];
						pDest->rgbBlue = pbSrc[0];
						pDest->rgbReserved  = 0x00;
					}
					pDest -= dwCX;
				}
			}
			else
			{
				for(j=0; j < dwCY; j++)
				{
					pDest -= dwCX;
					for(i=0; i < dwCX; i++, pbSrc+=4, pDest++)
					{
						pDest->rgbRed = pbSrc[0];
						pDest->rgbGreen = pbSrc[1];
						pDest->rgbBlue = pbSrc[2];
						pDest->rgbReserved  = 0x00;
					}
					pDest -= dwCX;
				}
			}
		} break;
	}
	pimgSrc->UnlockImage(ppixSrc);
	sz.cx = (int)dwCX;
	sz.cy = (int)dwCY;
	return hbm;
}

HBITMAP CreateErrorBitmap(IImage* pimgSrc, double errthresh, COLORREF clr, SIZE& sz)
{
	PixelFormat pf;
	pf.dwSize = sizeof(PixelFormat);
	DWORD dwCX, dwCY;
	pimgSrc->GetSize(&dwCX, &dwCY);
	pimgSrc->GetFormat(&pf);
	BITMAPINFO bi;
	RGBQUAD* prgbqPixels;
	RGBQUAD rgbqMask;
	rgbqMask.rgbBlue = GetBValue(clr) >> 1;
	rgbqMask.rgbGreen = GetGValue(clr) >> 1;
	rgbqMask.rgbRed = GetRValue(clr) >> 1;
	rgbqMask.rgbReserved = 0x00;
	void* ppixSrc;
	HDC hdc = ::GetDC(::GetDesktopWindow());
	bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bi.bmiHeader.biWidth = dwCX;
	bi.bmiHeader.biHeight = -(int)dwCY;
	bi.bmiHeader.biPlanes = 1;
	bi.bmiHeader.biBitCount = 32;
	bi.bmiHeader.biCompression = BI_RGB;
	bi.bmiHeader.biSizeImage = 0;
	bi.bmiHeader.biXPelsPerMeter = 640;
	bi.bmiHeader.biYPelsPerMeter = 640;
	bi.bmiHeader.biClrUsed = 0;
	bi.bmiHeader.biClrImportant = 0;
	HBITMAP hbm = ::CreateDIBSection(hdc, &bi, DIB_RGB_COLORS, (void**)&prgbqPixels, NULL, 0);
	if(!hbm)
		return NULL;
	::ReleaseDC(::GetDesktopWindow(), hdc);
	pimgSrc->LockImage(&ppixSrc);
	RGBQUAD* pDest = prgbqPixels + (dwCX * dwCY);
	DWORD i, j;
	switch(pf.dwTypeFlags)
	{
	case PF_FLOAT | PF_LONG:
		{
			double * pdSrc = (double*)ppixSrc;
			if(pf.bChannels == 1)
			{
				for(j=0; j < dwCY; j++)
				{
					pDest -= dwCX;
					for(i=0; i < dwCX; i++, pdSrc++, pDest++)
					{
						pDest->rgbReserved  = 0x00;
						if(*pdSrc <= errthresh)
						{
							pDest->rgbRed = (BYTE)(127.0 * pdSrc[0]) + rgbqMask.rgbRed;
							pDest->rgbGreen = (BYTE)(127.0 * pdSrc[0]) + rgbqMask.rgbGreen;
							pDest->rgbBlue = (BYTE)(127.0 * pdSrc[0]) + rgbqMask.rgbBlue;
						}
						else
						{
							pDest->rgbRed = (BYTE)(255.0 * pdSrc[0]);
							pDest->rgbGreen = (BYTE)(255.0 * pdSrc[0]);
							pDest->rgbBlue = (BYTE)(255.0 * pdSrc[0]);
						}
					}
					pDest -= dwCX;
				}
			}
		} break;
	case PF_INT | PF_BYTE :
		{
			BYTE * pbSrc = (BYTE*)ppixSrc;
			BYTE berr = (BYTE)(255.0 * errthresh); 
			if(pf.bChannels == 1)
			{
				for(j=0; j < dwCY; j++)
				{
					pDest -= dwCX;
					for(i=0; i < dwCX; i++, pbSrc++, pDest++)
					{
						pDest->rgbReserved  = 0x00;
						if(*pbSrc <= berr)
						{
							pDest->rgbRed = (pbSrc[0] >> 1) + rgbqMask.rgbRed;
							pDest->rgbGreen = (pbSrc[0] >> 1) + rgbqMask.rgbGreen;
							pDest->rgbBlue = (pbSrc[0] >> 1) + rgbqMask.rgbBlue;
						}
						else
						{
							pDest->rgbRed = pbSrc[0];
							pDest->rgbGreen = pbSrc[0];
							pDest->rgbBlue = pbSrc[0];
						}
					}
					pDest -= dwCX;
				}
			}
		} break;
	}
	pimgSrc->UnlockImage(ppixSrc);
	sz.cx = (int)dwCX;
	sz.cy = (int)dwCY;
	return hbm;
}

HBITMAP CreateMaskedBitmap(IImage* pimgSrc, IImage* pimgMask, COLORREF clr, SIZE& sz)
{
	PixelFormat pf;
	pf.dwSize = sizeof(PixelFormat);
	DWORD dwCX, dwCY;
	pimgSrc->GetSize(&dwCX, &dwCY);
	pimgSrc->GetFormat(&pf);
	BITMAPINFO bi;
	RGBQUAD* prgbqPixels;
	RGBQUAD rgbqMask;
	rgbqMask.rgbBlue = GetBValue(clr);
	rgbqMask.rgbGreen = GetGValue(clr);
	rgbqMask.rgbRed = GetRValue(clr);
	rgbqMask.rgbReserved = 0x00;
	void* ppixSrc, *ppixMask;
	HDC hdc = ::GetDC(::GetDesktopWindow());
	bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bi.bmiHeader.biWidth = dwCX;
	bi.bmiHeader.biHeight = -(int)dwCY;
	bi.bmiHeader.biPlanes = 1;
	bi.bmiHeader.biBitCount = 32;
	bi.bmiHeader.biCompression = BI_RGB;
	bi.bmiHeader.biSizeImage = 0;
	bi.bmiHeader.biXPelsPerMeter = 640;
	bi.bmiHeader.biYPelsPerMeter = 640;
	bi.bmiHeader.biClrUsed = 0;
	bi.bmiHeader.biClrImportant = 0;
	HBITMAP hbm = ::CreateDIBSection(hdc, &bi, DIB_RGB_COLORS, (void**)&prgbqPixels, NULL, 0);
	if(!hbm)
		return NULL;
	::ReleaseDC(::GetDesktopWindow(), hdc);
	pimgSrc->LockImage(&ppixSrc);
	pimgMask->LockImage(&ppixMask);
	RGBQUAD* pDest = prgbqPixels + (dwCX * dwCY);
	DWORD i, j;
	BYTE* pbMask = (BYTE*)ppixMask;
	switch(pf.dwTypeFlags)
	{
	case PF_FLOAT | PF_LONG:
		{
			double * pdSrc = (double*)ppixSrc;
			if(pf.bChannels == 1)
			{
				for(j=0; j < dwCY; j++)
				{
					pDest -= dwCX;
					for(i=0; i < dwCX; i++, pdSrc++, pDest++, pbMask++)
					{
						pDest->rgbReserved  = 0x00;
						if(*pbMask)
						{
							pDest->rgbRed = rgbqMask.rgbRed;
							pDest->rgbGreen = rgbqMask.rgbGreen;
							pDest->rgbBlue = rgbqMask.rgbBlue;
						}
						else
						{
							pDest->rgbRed = (BYTE)(255.0 * pdSrc[0]);
							pDest->rgbGreen = (BYTE)(255.0 * pdSrc[0]);
							pDest->rgbBlue = (BYTE)(255.0 * pdSrc[0]);
						}
					}
					pDest -= dwCX;
				}
			}
			else
			{
				for(j=0; j < dwCY; j++)
				{
					pDest -= dwCX;
					for(i=0; i < dwCX; i++, pdSrc+=4, pDest++, pbMask++)
					{
						pDest->rgbReserved  = 0x00;
						if(*pbMask)
						{
							pDest->rgbRed = rgbqMask.rgbRed;
							pDest->rgbGreen = rgbqMask.rgbGreen;
							pDest->rgbBlue = rgbqMask.rgbBlue;
						}
						else
						{
							pDest->rgbRed = (BYTE)(255.0 * pdSrc[0]);
							pDest->rgbGreen = (BYTE)(255.0 * pdSrc[1]);
							pDest->rgbBlue = (BYTE)(255.0 * pdSrc[2]);
						}
					}
					pDest -= dwCX;
				}
			}
		} break;
	case PF_INT | PF_BYTE :
		{
			BYTE * pbSrc = (BYTE*)ppixSrc;
			if(pf.bChannels == 1)
			{
				for(j=0; j < dwCY; j++)
				{
					pDest -= dwCX;
					for(i=0; i < dwCX; i++, pbSrc++, pDest++, pbMask++)
					{
						pDest->rgbReserved  = 0x00;
						if(*pbMask)
						{
							pDest->rgbRed = rgbqMask.rgbRed;
							pDest->rgbGreen = rgbqMask.rgbGreen;
							pDest->rgbBlue = rgbqMask.rgbBlue;
						}
						else
						{
							pDest->rgbRed = pbSrc[0];
							pDest->rgbGreen = pbSrc[0];
							pDest->rgbBlue = pbSrc[0];
						}
					}
					pDest -= dwCX;
				}
			}
			else
			{
				for(j=0; j < dwCY; j++)
				{
					pDest -= dwCX;
					for(i=0; i < dwCX; i++, pbSrc+=4, pDest++, pbMask++)
					{
						pDest->rgbReserved  = 0x00;
						if(*pbMask)
						{
							pDest->rgbRed = rgbqMask.rgbRed;
							pDest->rgbGreen = rgbqMask.rgbGreen;
							pDest->rgbBlue = rgbqMask.rgbBlue;
						}
						else
						{
							pDest->rgbRed = pbSrc[0];
							pDest->rgbGreen = pbSrc[1];
							pDest->rgbBlue = pbSrc[2];
						}
					}
					pDest -= dwCX;
				}
			}
		} break;
	}
	pimgSrc->UnlockImage(ppixSrc);
	pimgMask->UnlockImage(ppixMask);
	sz.cx = (int)dwCX;
	sz.cy = (int)dwCY;
	return hbm;
}

BOOL SaveBitmap(HBITMAP hbmp, LPCTSTR lpszFilename)
{
	PICTDESC pd;
	HRESULT hr;
	memset(&pd, 0, sizeof(PICTDESC));
	pd.cbSizeofstruct = sizeof(PICTDESC);
	pd.picType = PICTYPE_BITMAP;
	pd.bmp.hbitmap = hbmp;
	IPersistStream* pPic;
	IStream* pstm;
	hr = ::OleCreatePictureIndirect(&pd, IID_IPersistStream, FALSE, (void**)&pPic);
	if(FAILED(hr))
	{
		return FALSE;
	}
	hr = ::CreateStreamOnHGlobal(NULL, TRUE, &pstm);
	if(FAILED(hr))
	{
		pPic->Release();
	}
	hr = pPic->Save(pstm, FALSE);
	pPic->Release();
	if(FAILED(hr))
	{
		pstm->Release();
		return FALSE;
	}
	STATSTG stat;
	pstm->Stat(&stat, STATFLAG_NONAME);
	HGLOBAL hg;
	GetHGlobalFromStream(pstm, &hg);
	HANDLE hfile;
	hfile = ::CreateFile(lpszFilename, GENERIC_WRITE, FILE_SHARE_READ, NULL, 
		CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if(hfile)
	{
		BYTE* p = (BYTE*)GlobalLock(hg);
		p += 8; //some header crap
		DWORD dw;
		if(!::WriteFile(hfile, p, stat.cbSize.LowPart, &dw, NULL))
			hr = HRESULT_FROM_WIN32(GetLastError());
		GlobalUnlock(hg);
		CloseHandle(hfile);
	}
	else
		hr = HRESULT_FROM_WIN32(GetLastError());
	pstm->Release();
	return FALSE;
}
