/**
 *  Filters for the white light method in Textor
 *
 **/

#include "imgproc.h"
#include <math.h>
#include "trans.h"

HRESULT __stdcall CoUnitVImage(IImage* pimg)
{
	DWORD dwCX, dwCY;
	PixelFormat pf;
	pf.dwSize = sizeof(PixelFormat);
	void* ppix;
	HRESULT hr;
	if(FAILED(hr = pimg->GetSize(&dwCX, &dwCY)))
		return hr;
	if(FAILED(hr = pimg->GetFormat(&pf)))
		return hr;
	if(pf.bChannels < 3 || (pf.dwTypeFlags != (PF_FLOAT | PF_LONG)))
		return IMAGE_E_PIXELFORMAT;
	if(FAILED(hr = pimg->LockImage(&ppix)))
		return hr;
	double* pDest = (double*)ppix;
	double register acc;
	for(DWORD j=0; j < dwCY; j++)
		for(DWORD i=0; i < dwCX; i++, pDest+=pf.bChannels)
		{
			acc = pDest[0] * pDest[0];
			acc += pDest[1] * pDest[1];
			acc += pDest[2] * pDest[2];
			acc = sqrt(acc);
			if(acc > 0.0)
			{
				double register accinv = 1.0 / acc; 
				pDest[0] *= accinv;
				pDest[1] *= accinv;
				pDest[2] *= accinv;
			}
		}
	pimg->UnlockImage(ppix);
	return S_OK;
}

HRESULT __stdcall CoCalcOccError(IImage* pimgSrc, IImage* pimgAveNorm, IImage* pimgResult)
{
	DWORD dwCX, dwCY;
	void* ppixSrc, *ppixAveNorm, *ppixResult;
	pimgSrc->GetSize(&dwCX, &dwCY);
	pimgSrc->LockImage(&ppixSrc);
	pimgAveNorm->LockImage(&ppixAveNorm);
	pimgResult->LockImage(&ppixResult);
	double* pSrc = (double*)ppixSrc;
	double* pAveNorm = (double*)ppixAveNorm;
	double* pResult = (double*)ppixResult;
	for(DWORD j = 0; j < dwCY; j++)
		for(DWORD i=0; i < dwCX; i++, pSrc+=4, pAveNorm+=4, pResult++)
		{
			*pResult = DotP(pSrc, pAveNorm);
		}
	pimgSrc->UnlockImage(&ppixSrc);
	pimgAveNorm->UnlockImage(&ppixAveNorm);
	pimgResult->UnlockImage(&ppixResult);
	return S_OK;
}

HRESULT __stdcall CoGrayScaleImage(IImage* pimgSrc, IImage* pimgRes)
{
	DWORD dwCX, dwCY;
	double factor = 1.0 / 3.0;
	pimgSrc->GetSize(&dwCX, &dwCY);
	void* ppixSrc, *ppixRes;
	pimgSrc->LockImage(&ppixSrc);
	pimgRes->LockImage(&ppixRes);
	double* pSrc = (double*)ppixSrc;
	double* pRes = (double*)ppixRes;
	for(DWORD j=0; j < dwCY; j++)
		for(DWORD i=0; i < dwCX; i++, pSrc+=4, pRes++)
		{
			*pRes = pSrc[0] + pSrc[1] + pSrc[2];
			*pRes *= factor;
		}
	pimgSrc->UnlockImage(ppixSrc);
	pimgRes->UnlockImage(ppixRes);
	return S_OK;
}

HRESULT __stdcall CoAbsImage(IImage* pimgSrc, IImage* pimgRes)
{
	DWORD dwCX, dwCY;
	double factor = 1.0 / 3.0;
	pimgSrc->GetSize(&dwCX, &dwCY);
	void* ppixSrc, *ppixRes;
	pimgSrc->LockImage(&ppixSrc);
	pimgRes->LockImage(&ppixRes);
	double* pSrc = (double*)ppixSrc;
	double* pRes = (double*)ppixRes;
	for(DWORD j=0; j < dwCY; j++)
		for(DWORD i=0; i < dwCX; i++, pSrc+=4, pRes++)
		{
			*pRes =  pSrc[0] * pSrc[0];
			*pRes += pSrc[1] * pSrc[1];
			*pRes += pSrc[2] * pSrc[2];
			*pRes = sqrt(*pRes);
		}
	pimgSrc->UnlockImage(ppixSrc);
	pimgRes->UnlockImage(ppixRes);
	return S_OK;
}


HRESULT __stdcall CoDivideSingleChannelImage(IImage* pimgSrc, IImage* pimgDenom)
{
	DWORD dwCX, dwCY;
	double factor = 1.0 / 3.0;
	pimgSrc->GetSize(&dwCX, &dwCY);
	void* ppixSrc, *ppixDenom;
	pimgSrc->LockImage(&ppixSrc);
	pimgDenom->LockImage(&ppixDenom);
	double* pSrc = (double*)ppixSrc;
	double* pDenom = (double*)ppixDenom;
	for(DWORD j=0; j < dwCY; j++)
		for(DWORD i=0; i < dwCX; i++, pSrc++, pDenom++)
		{
			*pSrc /= *pDenom;
		}
	pimgSrc->UnlockImage(ppixSrc);
	pimgDenom->UnlockImage(ppixDenom);
	return S_OK;
}