/**
 *	Jon Rowlett
 *
 **/

#ifndef __IMAGEPROC_H__
#define __IMAGEPROC_H__

#include <Image.h>

HRESULT __stdcall CoCreateImage(IMalloc* pMalloc, DWORD dwCX, DWORD dwCY, 
								PixelFormat* pPixelFormat, REFIID riid, void** ppv);
HRESULT __stdcall CoWarpImage(IMalloc* pMalloc, IImage* pSrc, double ptSrc[8], DWORD dwCX, DWORD dwCY, IImage** ppOut);
HRESULT __stdcall CoCopyImage(IMalloc* pMalloc, IImage* pSrc, PixelFormat* pNewFormat, IImage** ppOut);

HRESULT __stdcall CoExtractTextels(IMalloc* pMalloc, IImage* pSrc, double ptSrc[8], 
								   int nxdiv, int nydiv, DWORD dwCX, DWORD dwCY, IImage** ppOut);

HRESULT __stdcall CoBltPerspective(IImage* pimgSrc, double ptDest[8], IImage* pDest);

HRESULT __stdcall SaveImage(IImage* pimgSrc, LPCTSTR lpszFileName);

/**
 *  Computes the minimum resolution of a square to be extracted from the points
 *  See notes 11.25.00
 **/
void ComputeMinRes(double pts[8], double &cx, double &cy);

/**
 *  Computes average texture from several texture images
 *  See notes 11.25.01
 *  can return E_OUTOFMEMORY and IMAGE_E_DIMENSION
 */
HRESULT __stdcall CoAverageImage(IMalloc* pMalloc, IImage* pimgSrc, int nxdiv, int nydiv, IImage** ppimgOut);

HRESULT __stdcall CoCorrectPerspective(IMalloc* pMalloc, IImage* pimgSrc, double pts[8], int nxdiv, int nydiv, IImage** ppimgOut);

HRESULT __stdcall CoRepeatImage(IMalloc* pMalloc, IImage* pimgSrc, int nxdiv, int nydiv, IImage** ppimg);

typedef void (*IMGFUNC)(DWORD x, DWORD y, double src1[4], double src2[4], double dest[4]);

HRESULT __stdcall ApplyFunction(IImage* pimgSrc1, IImage* pimgSrc2, IImage* pimgDest, IMGFUNC pFunc);

HRESULT __stdcall ResampleImage(IMalloc* pMalloc, IImage* pimgSrc, DWORD dwCX, DWORD dwCY, IImage** ppimgOut);

/**
 *  Interoperability with Garland / Heckbert Terra System
 *  PGM support
 *  11/03/1999
 **/

HRESULT __stdcall SaveChannelPGM(IImage* pimgSrc, BYTE bChannel, LPCTSTR lpszFilename, bool bText = true);

HRESULT __stdcall SavePGM(IImage* pimgSrc, LPCTSTR lpszFilename, bool bText = true);
HRESULT __stdcall LoadTIN(LPCTSTR lpszFilename, IImageFactory* pimgf, IImage** ppimgOut);
HRESULT __stdcall CompositeRGB(IImage* pimgR, IImage* pimgG, IImage* pimgB, IImageFactory* pimgf, IImage** ppimgOut);

IImageFactory* __stdcall GetHeapImageFactory();

/** 
 *	computes median texture from several texture images
 **/

HRESULT __stdcall CoMinImage(IImageFactory* pimgf, IImage* pimgSrc, int nxdiv, int nydiv, IImage** ppimgOut);
HRESULT __stdcall CoMaxImage(IImageFactory* pimgf, IImage* pimgSrc, int nxdiv, int nydiv, IImage** ppimgOut);
HRESULT __stdcall CoMedianImage(IImageFactory* pimgf, IImage* pimgMin, IImage* pimgMax, IImage** ppimgOut);

// 11/22/99 
// real median.  Other median is really mode
HRESULT __stdcall CoMedianImage(IImageFactory* pimgf, IImage* pimgSrc, int nxdiv, int nydiv, IImage** ppimgOut);
HRESULT __stdcall CoStddevImage(IImage* pimgSrc, IImage* pimgAve, int nxdiv, int nydiv, IImageFactory* pimgf, IImage** ppimgOut);

// for 869 project

void __stdcall SquareDiff(double* pimg1, double* pimg2, int cx1, int cx2, int x1, int y1, int x2, int y2, int w, double *res);
double __stdcall SumSquareDiff(int w, double* res);

double SSD(double* pSrc, double* pTest, 
		   int cxSrc, int cySrc,
		   int cxTest, int cyTest,
		   int xSrc, int ySrc,
		   int xTest, int yTest, int w, int* factor);

double SSDEx(double* pSrc, double* pTest, BYTE* pbMask,
			 int cxSrc, int cySrc,
			 int cxTest, int cyTest,
			 int xSrc, int ySrc,
			 int xTest, int yTest, int w, int* factor);

HRESULT __stdcall FindMaskEdge(IImage* pimgSrc, IImage* pimgDest);

double __stdcall FindAlpha(double* prgbBack, double* prgbFront, double* prgbTrue);
HRESULT __stdcall SynthTex(IImage* pimgSrc, IImage* pimgEdge, 
						   IImage* pimgFlowX, IImage* pimgFlowY, IImage* pimgTexSynth);
HRESULT __stdcall SynthOcc(IImage* pimgSrc, IImage* pimgTexMask, IImage* pimgEdge, 
						   IImage* pimgOccSynth);
HRESULT __stdcall ExtractAlpha(IImage* pimgSrc, IImage* pimgTexMask, IImage* pimgEdge, 
							   IImage* pimgTexSynth, IImage* pimgOccSynth, 
							   IImage* pimgAlphaMask);

/**
 *  White light method
 **/

HRESULT __stdcall CoUnitVImage(IImage* pimg);
HRESULT __stdcall CoCalcOccError(IImage* pimgSrc, IImage* pimgAveNorm, IImage* pimgResult);
HRESULT __stdcall CoGrayScaleImage(IImage* pimgSrc, IImage* pimgRes);
HRESULT __stdcall CoAbsImage(IImage* pimgSrc, IImage* pimgRes);
HRESULT __stdcall CoDivideSingleChannelImage(IImage* pimgSrc, IImage* pimgDenom);
HRESULT __stdcall CoInvertMask(IImage* pimg);
HRESULT __stdcall CoThresholdError(IImage* pimg, double thresh, IImageFactory* pimgf, IImage** ppimgMask);
HRESULT __stdcall CoBltMask(IImage* pimgSrc, IImage* pimgMask, IImage* pimgDest); 
HRESULT __stdcall CreateColorImage(IImageFactory* pimgf, DWORD dwCX, DWORD dwCY, double r, double b, double g, double a, IImage** ppimgOut);
HRESULT __stdcall ApplyShadow(IImage* pimgDest, IImage* pimgShadow);

#endif /* __IMAGEPROC_H__ */