#include "imgproc.h"

HRESULT __stdcall CoMinImage(IImageFactory* pimgf, IImage* pimgSrc, int nxdiv, int nydiv, IImage** ppimgOut)
{
	DWORD dwSrcCX, dwSrcCY;
	DWORD dwCX, dwCY;
	PixelFormat pf;
	pf.dwSize = sizeof(PixelFormat);
	pf.dwTypeFlags = PF_FLOAT | PF_LONG;
	pf.bChannels = 4;
	HRESULT hr = pimgSrc->GetSize(&dwSrcCX, &dwSrcCY);
	if(FAILED(hr))
		return hr;
	dwCX = dwSrcCX / nxdiv;
	dwCY = dwSrcCY / nydiv;
	if(dwCX * nxdiv != dwSrcCX || dwCY * nydiv != dwSrcCY)
		return IMAGE_E_DIMENSION;
	IImage* pimg;
	hr = pimgf->CreateImage(dwCX, dwCY, &pf, &pimg);
	if(FAILED(hr))
		return hr;
	void* ppixDest;
	void* ppixSrc;
	pimgSrc->LockImage(&ppixSrc);
	pimg->LockImage(&ppixDest);
	double* pDest;
	double* pTex;
	double* pSrc;
	DWORD i, j, k, l;
	DWORD dRow = (dwSrcCX - dwCX) * 4;
	DWORD dTexRow = (dwCY * dwSrcCX - dwSrcCX) * 4;
	pDest = (double*)ppixDest;
	pTex = (double*)ppixSrc;
	pSrc = pTex;
	for(j = 0; j < dwCY; j++, pSrc+=dRow)
		for(i=0; i < dwCX; i++, pDest+=4, pSrc+=4)
		{
			pDest[0] = pSrc[0];
			pDest[1] = pSrc[1];
			pDest[2] = pSrc[2];
			pDest[3] = pSrc[3];
		}
	for(l = 0, pTex = (double*)ppixSrc; l < (DWORD)nydiv; l++, pTex += dTexRow)
		for(k = 0; k < (DWORD)nxdiv; k++, pTex += (4 * dwCX))
		{
			pDest = (double*)ppixDest;
			pSrc = pTex;
			for(j = 0; j < dwCY; j++, pSrc+=dRow)
				for(i=0; i < dwCX; i++, pDest+=4, pSrc+=4)
				{
					pDest[0] = min(pDest[0], pSrc[0]);
					pDest[1] = min(pDest[1], pSrc[1]);
					pDest[2] = min(pDest[2], pSrc[2]);
					pDest[3] = min(pDest[3], pSrc[3]);
				}
		}
	pimgSrc->UnlockImage(ppixSrc);
	pimg->UnlockImage(ppixDest);
	*ppimgOut = pimg;
	return S_OK;
}

HRESULT __stdcall CoMaxImage(IImageFactory* pimgf, IImage* pimgSrc, int nxdiv, int nydiv, IImage** ppimgOut)
{
	DWORD dwSrcCX, dwSrcCY;
	DWORD dwCX, dwCY;
	PixelFormat pf;
	pf.dwSize = sizeof(PixelFormat);
	pf.dwTypeFlags = PF_FLOAT | PF_LONG;
	pf.bChannels = 4;
	HRESULT hr = pimgSrc->GetSize(&dwSrcCX, &dwSrcCY);
	if(FAILED(hr))
		return hr;
	dwCX = dwSrcCX / nxdiv;
	dwCY = dwSrcCY / nydiv;
	if(dwCX * nxdiv != dwSrcCX || dwCY * nydiv != dwSrcCY)
		return IMAGE_E_DIMENSION;
	IImage* pimg;
	hr = pimgf->CreateImage(dwCX, dwCY, &pf, &pimg);
	if(FAILED(hr))
		return hr;
	void* ppixDest;
	void* ppixSrc;
	pimgSrc->LockImage(&ppixSrc);
	pimg->LockImage(&ppixDest);
	memset(ppixDest, 0, sizeof(double) * 4 * dwCX * dwCY);
	double* pDest;
	double* pTex;
	double* pSrc;
	DWORD i, j, k, l;
	DWORD dRow = (dwSrcCX - dwCX) * 4;
	DWORD dTexRow = (dwCY * dwSrcCX - dwSrcCX) * 4;
	for(l = 0, pTex = (double*)ppixSrc; l < (DWORD)nydiv; l++, pTex += dTexRow)
		for(k = 0; k < (DWORD)nxdiv; k++, pTex += (4 * dwCX))
		{
			pDest = (double*)ppixDest;
			pSrc = pTex;
			for(j = 0; j < dwCY; j++, pSrc+=dRow)
				for(i=0; i < dwCX; i++, pDest+=4, pSrc+=4)
				{
					pDest[0] = max(pDest[0], pSrc[0]);
					pDest[1] = max(pDest[1], pSrc[1]);
					pDest[2] = max(pDest[2], pSrc[2]);
					pDest[3] = max(pDest[3], pSrc[3]);
				}
		}
	pimgSrc->UnlockImage(ppixSrc);
	pimg->UnlockImage(ppixDest);
	*ppimgOut = pimg;
	return S_OK;
}

HRESULT __stdcall CoMedianImage(IImageFactory* pimgf, IImage* pimgMin, IImage* pimgMax, IImage** ppimgOut)
{
	PixelFormat pf;
	pf.dwSize = sizeof(PixelFormat);
	pf.dwTypeFlags = PF_FLOAT | PF_LONG;
	pf.bChannels = 4;
	DWORD dwCX, dwCY;
	pimgMin->GetSize(&dwCX, &dwCY);
	pimgf->CreateImage(dwCX, dwCY, &pf, ppimgOut);
	void* ppixDest;
	void* ppixMin;
	void* ppixMax;
	pimgMin->LockImage(&ppixMin);
	pimgMax->LockImage(&ppixMax);
	(*ppimgOut)->LockImage(&ppixDest);
	double *pDest = (double*)ppixDest;
	double *pMin = (double*)ppixMin;
	double *pMax = (double*)ppixMax;
	for(DWORD j=0; j < dwCY; j++)
		for(DWORD i=0; i < dwCX; i++, pDest+=4, pMin+=4, pMax+=4)
		{
			pDest[0] = (pMin[0] + pMax[0]) * 0.5;
			pDest[1] = (pMin[1] + pMax[1]) * 0.5;
			pDest[2] = (pMin[2] + pMax[2]) * 0.5;
			pDest[3] = (pMin[3] + pMax[3]) * 0.5;
		}
	pimgMin->UnlockImage(ppixMin);
	pimgMax->UnlockImage(ppixMax);
	(*ppimgOut)->UnlockImage(ppixDest);
	return S_OK;
}
