#include "imgproc.h"
#include "texsyn.h"
#include <malloc.h>
#include <assert.h>

CPixelTestIterator::CPixelTestIterator(IImageFactory* pimgf, IImage* pimg)
: m_pimgMask(pimg), m_pimgf(pimgf), m_pimgNewMask(NULL)
{
	assert(pimg);
	assert(pimgf);
	m_pimgf->AddRef();
	m_pimgMask->AddRef();
	m_pimgMask->LockImage(&m_pPix);
	m_pbLoc = (BYTE*)m_pPix;
	m_pimgMask->GetSize(&m_dwCX, &m_dwCY);
	PixelFormat pf;
	m_pimgMask->GetFormat(&pf);
	m_pimgf->CreateImage(m_dwCX, m_dwCY, &pf,  &m_pimgNewMask);
	m_pimgNewMask->LockImage(&m_pNewPix);
	m_pbNewLoc = (BYTE*)m_pNewPix;
	memcpy(m_pNewPix, m_pPix, m_dwCX * m_dwCY);
	m_nX = 0;
	m_nY = 0;
	m_nPixelsTested = 0;
}

CPixelTestIterator::~CPixelTestIterator()
{
	m_pimgf->Release();
	m_pimgNewMask->UnlockImage(m_pNewPix);
	m_pimgNewMask->Release();
	m_pimgMask->UnlockImage(m_pPix);
	m_pimgMask->Release();
}

CPixelTestIterator& CPixelTestIterator::operator++()
{
	assert(m_pimgMask);
	*m_pbNewLoc = 0xFF;
	bool done;
	do
	{
		done = false;
		m_pbLoc++;
		m_pbNewLoc++;
		m_nX++;
		if(m_nX == (int)m_dwCX)
		{
			m_nX = 0;
			m_nY++;
			if(m_nY == (int)m_dwCY)
				break;
		}
		if(*m_pbLoc)
			continue;
		if(m_nX < (int)m_dwCX - 1 && m_pbLoc[1])
			break;
		if(m_nY < (int)m_dwCY - 1 && m_pbLoc[m_dwCX])
			break;
		if(m_nX > 0 && m_pbLoc[-1])
			break;
		if(m_nY > 0 && m_pbLoc[-(int)m_dwCX])
			break;
	} while(!done);
	m_nPixelsTested++;
	return *this;
}

void CPixelTestIterator::GetTestPixel(int& x, int& y)
{
	assert(m_pimgMask);
	x = m_nX;
	y = m_nY;
}

CPixelTestIterator::operator bool() const
{
	assert(m_pimgMask);
	return m_nY < (int)m_dwCY;
}

void CPixelTestIterator::begin()
{
	m_nX = 0;
	m_nY = 0;
	m_nPixelsTested = 0;
	m_pbLoc = (BYTE*)m_pPix;
	m_pbNewLoc = (BYTE*)m_pNewPix;
	while(m_nY < (int)m_dwCY)
	{
		if(!*m_pbLoc)
		{
			if(m_nX < (int)m_dwCX - 1 && m_pbLoc[1])
				break;
			if(m_nY < (int)m_dwCY - 1 && m_pbLoc[m_dwCX])
				break;
			if(m_nX > 0 && m_pbLoc[-1])
				break;
			if(m_nY > 0 && m_pbLoc[-(int)m_dwCX])
				break;
		}
		m_nX++;
		m_pbNewLoc++;
		m_pbLoc++;
		if(m_nX == (int)m_dwCX)
		{
			m_nY++;
			m_nX = 0;
		}
	}
}

int CPixelTestIterator::GetPixelsTested()
{
	return m_nPixelsTested;
}

void CPixelTestIterator::GetNewMask(IImage** ppimgMask)
{
	*ppimgMask = m_pimgNewMask;
	m_pimgNewMask->AddRef();
}

CTexWindowIterator::CTexWindowIterator(IImage* pimg, int w)
: m_nSize(w), m_nX(0), m_nY(0)
{
	pimg->GetSize(&m_dwCX, &m_dwCY);
}

CTexWindowIterator::~CTexWindowIterator()
{
}

// retrieves the coords of the texture window
void CTexWindowIterator::getCoords(int& x, int& y)
{
	x = m_nX;
	y = m_nY;
}

CRectTexWindowIterator::CRectTexWindowIterator(IImage* pimg, int w, int x, int y, int cx, int cy)
: CTexWindowIterator(pimg, w), m_nX0(x), m_nY0(y), m_nX1(x + cx), m_nY1(y + cy)
{
}

#ifdef WIN32
CRectTexWindowIterator::CRectTexWindowIterator(IImage* pimg, int w, const RECT& r)
: CTexWindowIterator(pimg, w), m_nX0(r.left), m_nY0(r.top), m_nX1(r.right), m_nY1(r.bottom)
{
}
#endif

// resets the iterator
void CRectTexWindowIterator::begin()
{
	m_nX = m_nX0 + m_nSize;
	m_nY = m_nY0 + m_nSize;
}

// tells when the iterator has finished and the coords are no longer valid
bool CRectTexWindowIterator::end()
{
	return m_nY > m_nY1 - m_nSize;
}

// finds next texture window
CTexWindowIterator& CRectTexWindowIterator::inc()
{
	m_nX++;
	if(m_nX > m_nX1 - m_nSize)
	{
		m_nY++;
		m_nX = m_nX0 + m_nSize;
	}
	return *this;
}


void __stdcall SquareDiff(double* pimg1, double* pimg2, int cx1, int cx2, int x1, int y1, int x2, int y2, int w, double *res)
{
	x1 -= w;
	x2 -= w;
	y1 -= w;
	y2 -= w;
	int i, j;
	double accum;
	pimg1 += (cx1 * y1 + x1) * 4;
	pimg2 += (cx2 * y2 + x2) * 4;
	int dp1 = (cx1 - 2 * w - 1) * 4;
	int dp2 = (cx2 - 2 * w - 1) * 4;
	for(j = 0; j <= 2 * w; j++, pimg1 += dp1, pimg2 += dp2)
		for(i = 0; i <= 2 * w; i++, pimg1 += 4, pimg2 += 4, res++)
		{
			accum = (pimg1[0] - pimg2[0]);
			accum *= accum;
			*res = accum;
			accum = (pimg1[1] - pimg2[1]);
			accum *= accum;
			*res += accum;
			accum = (pimg1[2] - pimg2[2]);
			accum *= accum;
			*res += accum;
		}
}

double __stdcall SumSquareDiff(int w, double* res)
{
	double accum = *res;
	int end = 2* w +1;
	end *= end;
	for(int i=1; i < end; i++)
		accum += res[i];
	return accum;
}

double SSD(double* pSrc, double* pTest, 
		   int cxSrc, int cySrc,
		   int cxTest, int cyTest,
		   int xSrc, int ySrc,
		   int xTest, int yTest, int w, int* factor)
{
	double ans = 0;
	double accum;
	int sx = max(max(-xSrc,0),max(-xTest,0));
	int sx1 = xSrc + sx;
	int sx2 = xTest + sx;
	int sy = max(max(-ySrc,0),max(-yTest,0));
	int sy1 = ySrc + sy;
	int sy2 = yTest + sy;
	int wx1 = min(cxSrc - xSrc, w) - max(-xSrc,0);
	int wx2 = min(cxTest - xTest, w) - max(-xTest,0);
	int wx = min(wx1, wx2);
	int wy1 = min(cySrc - ySrc, w) - max(-ySrc,0);
	int wy2 = min(cyTest - yTest, w) - max(-yTest,0);
	int wy = min(wy1, wy2);
	int dy1 = (cxSrc - wx1) << 2;
	int dy2 = (cxTest - wx2) << 2;
	pSrc += (cxSrc * sy1 + sx1) << 2;
	pTest += (cxTest * sy2 + sx2) << 2;
	for(int j=0; j < wy; j++, pSrc+=dy1, pTest+=dy2)
		for(int i=0; i < wx; i++, pSrc+=4, pTest+=4)
		{
			accum = (pSrc[0] - pTest[0]);
			accum *= accum;
			ans += accum;
			accum = (pSrc[1] - pTest[1]);
			accum *= accum;
			ans += accum;
			accum = (pSrc[2] - pTest[2]);
			accum *= accum;
			ans += accum;
		}
	*factor = wx * wy;
	return ans;
}

double SSDEx(double* pSrc, double* pTest, BYTE* pbMask,
			 int cxSrc, int cySrc,
			 int cxTest, int cyTest,
			 int xSrc, int ySrc,
			 int xTest, int yTest, int w, int* factor)
{
	double ans = 0;
	double accum;
	*factor = 0;
	int sx = max(max(-xSrc,0),max(-xTest,0));
	int sx1 = xSrc + sx;
	int sx2 = xTest + sx;
	int sy = max(max(-ySrc,0),max(-yTest,0));
	int sy1 = ySrc + sy;
	int sy2 = yTest + sy;
	int wx1 = min(cxSrc - xSrc, w) - max(-xSrc,0);
	int wx2 = min(cxTest - xTest, w) - max(-xTest,0);
	int wx = min(wx1, wx2);
	int wy1 = min(cySrc - ySrc, w) - max(-ySrc,0);
	int wy2 = min(cyTest - yTest, w) - max(-yTest,0);
	int wy = min(wy1, wy2);
	int dy1 = (cxSrc - wx1) << 2;
	int dy3 = (cxTest - wx2);
	int dy2 =  dy3 << 2;
	pSrc += (cxSrc * sy1 + sx1) << 2;
	pTest += (cxTest * sy2 + sx2) << 2;
	pbMask += (cxTest * sy2 + sx2);
	for(int j=0; j < wy; j++, pSrc+=dy1, pTest+=dy2, pbMask+=dy3)
		for(int i=0; i < wx; i++, pSrc+=4, pTest+=4, pbMask++)
		{
			if(*pbMask)
			{
				accum = (pSrc[0] - pTest[0]);
				accum *= accum;
				ans += accum;
				accum = (pSrc[1] - pTest[1]);
				accum *= accum;
				ans += accum;
				accum = (pSrc[2] - pTest[2]);
				accum *= accum;
				ans += accum;
				*factor++;
			}
		}
	return ans;
}

double __stdcall FindAlpha(double* prgbBack, double* prgbFront, double* prgbTrue)
{
	double a[3];
	a[0] = (prgbTrue[0] - prgbBack[0]) / (prgbFront[0] - prgbBack[0]);
	a[1] = (prgbTrue[1] - prgbBack[1]) / (prgbFront[1] - prgbBack[1]);
	a[2] = (prgbTrue[2] - prgbBack[2]) / (prgbFront[2] - prgbBack[2]);
	if(a[0] < 0.0 || a[0] > 1.0)
		a[0] = 1.0;
	if(a[1] < 0.0 || a[1] > 1.0)
		a[1] = 1.0;
	if(a[2] < 0.0 || a[2] > 1.0)
		a[2] = 1.0;
	return (a[0] + a[1] + a[2]) / 3.0;
}

HRESULT __stdcall SynthTex(IImage* pimgSrc, IImage* pimgEdge, 
						   IImage* pimgFlowX, IImage* pimgFlowY, IImage* pimgTexSynth)
{
	DWORD dwCX, dwCY;
	pimgSrc->GetSize(&dwCX, &dwCY);
	void* ppixSrc, *ppixEdge, *ppixFlowX, *ppixFlowY, *ppixTexSynth;
	double* pdSrc, *pdTexSynth;
	BYTE* pbEdge;
	int* pnFlowX, *pnFlowY;
	pimgSrc->LockImage(&ppixSrc);
	pimgEdge->LockImage(&ppixEdge);
	pimgTexSynth->LockImage(&ppixTexSynth);
	pimgFlowX->LockImage(&ppixFlowX);
	pimgFlowY->LockImage(&ppixFlowY);
	pdSrc = (double*)ppixSrc;
	pdTexSynth = (double*)ppixTexSynth;
	pbEdge = (BYTE*)ppixEdge;
	pnFlowX = (int*)ppixFlowX;
	pnFlowY = (int*)ppixFlowY;
	for(DWORD j=0; j < dwCY; j++)
		for(DWORD i=0; i < dwCX; i++, pbEdge++, pnFlowX++, pnFlowY++, pdTexSynth+=4)
		{
			if(*pbEdge)
			{
				memcpy(pdTexSynth, pdSrc + dwCX* *pnFlowY + *pnFlowX, sizeof(double) * 4);
			}
		}
	pimgSrc->UnlockImage(ppixSrc);
	pimgEdge->UnlockImage(ppixEdge);
	pimgTexSynth->UnlockImage(ppixTexSynth);
	pimgFlowX->UnlockImage(ppixFlowX);
	pimgFlowY->UnlockImage(ppixFlowY);
	return S_OK;
}

HRESULT __stdcall SynthOcc(IImage* pimgSrc, IImage* pimgTexMask, IImage* pimgEdge, 
						   IImage* pimgOccSynth)
{
	static double factor[5] = {0.0, 1.0, 0.5, 0.3333, 0.25};
	DWORD dwCX, dwCY;
	void* ppixSrc, *ppixTexMask, *ppixEdge, *ppixOccSynth;
	double* pdSrc, *pdOccSynth;
	BYTE* pbTexMask, *pbEdge;
	pimgSrc->GetSize(&dwCX, &dwCY);
	pimgSrc->LockImage(&ppixSrc);
	pimgTexMask->LockImage(&ppixTexMask);
	pimgEdge->LockImage(&ppixEdge);
	pimgOccSynth->LockImage(&ppixOccSynth);
	pdSrc = (double*)ppixSrc;
	pdOccSynth = (double*)ppixOccSynth;
	pbTexMask = (BYTE*)ppixTexMask;
	pbEdge = (BYTE*)ppixEdge;
	int n = 0;
	for(DWORD j=0; j < dwCY; j++)
		for(DWORD i=0; i < dwCX; i++, pbEdge++, pbTexMask, pdOccSynth+=4, pdSrc+=4)
		{
			if(*pbEdge)
			{
				n = 0;
				if(i < dwCX-1 && !*(pbTexMask+1))
				{
					pdOccSynth[0] = (pdSrc+4)[0];
					pdOccSynth[1] = (pdSrc+4)[1];
					pdOccSynth[2] = (pdSrc+4)[2];
					n++;
				}
				if(i > 0 && !*(pbTexMask-1))
				{
					pdOccSynth[0] += (pdSrc-4)[0];
					pdOccSynth[1] += (pdSrc-4)[1];
					pdOccSynth[2] += (pdSrc-4)[2];
					n++;
				}
				if(j < dwCY-1 && !*(pbTexMask+dwCX))
				{
					pdOccSynth[0] += (pdSrc+(dwCX << 2))[0];
					pdOccSynth[1] += (pdSrc+(dwCX << 2))[1];
					pdOccSynth[2] += (pdSrc+(dwCX << 2))[2];
					n++;
				}
				if(j > 0 && !*(pbTexMask-dwCX))
				{
					pdOccSynth[0] += (pdSrc-(dwCX << 2))[0];
					pdOccSynth[1] += (pdSrc-(dwCX << 2))[1];
					pdOccSynth[2] += (pdSrc-(dwCX << 2))[2];
					n++;
				}
				pdOccSynth[0] *= factor[n];
				pdOccSynth[1] *= factor[n];
				pdOccSynth[2] *= factor[n];
			}
		}
	pimgSrc->UnlockImage(ppixSrc);
	pimgTexMask->UnlockImage(ppixTexMask);
	pimgEdge->UnlockImage(ppixEdge);
	pimgOccSynth->UnlockImage(ppixOccSynth);
	return S_OK;
}

HRESULT __stdcall ExtractAlpha(IImage* pimgSrc, IImage* pimgTexMask, IImage* pimgEdge, 
							   IImage* pimgTexSynth, IImage* pimgOccSynth, 
							   IImage* pimgAlphaMask)
{
	DWORD dwCX, dwCY;
	void* ppixSrc, *ppixTexMask, *ppixEdge, *ppixTexSynth, *ppixOccSynth, *ppixAlphaMask;
	double* pdSrc, *pdTexSynth, *pdOccSynth, *pdAlphaMask;
	BYTE* pbTexMask, *pbEdge;
	pimgSrc->GetSize(&dwCX, &dwCY);
	pimgSrc->LockImage(&ppixSrc);
	pimgTexMask->LockImage(&ppixTexMask);
	pimgEdge->LockImage(&ppixEdge);
	pimgTexSynth->LockImage(&ppixTexSynth);
	pimgOccSynth->LockImage(&ppixOccSynth);
	pimgAlphaMask->LockImage(&ppixAlphaMask);
	pdSrc = (double*)ppixSrc;
	pdTexSynth = (double*)ppixTexSynth;
	pdOccSynth = (double*)ppixOccSynth;
	pdAlphaMask = (double*)ppixAlphaMask;
	pbTexMask = (BYTE*)ppixTexMask;
	pbEdge = (BYTE*)ppixEdge;
	for(DWORD j=0; j < dwCY; j++)
		for(DWORD i=0; i < dwCX; i++, pdSrc+=4, pdTexSynth+=4, 
			pdOccSynth+=4, pdAlphaMask++, pbTexMask++, pbEdge++)
		{
			if(*pbEdge)
				*pdAlphaMask = FindAlpha(pdTexSynth, pdOccSynth, pdSrc);
			else
				*pdAlphaMask = *pbTexMask ? 0.0 : 1.0;
		}
	pimgSrc->UnlockImage(ppixSrc);
	pimgTexMask->UnlockImage(ppixTexMask);
	pimgEdge->UnlockImage(ppixEdge);
	pimgTexSynth->UnlockImage(ppixTexSynth);
	pimgOccSynth->UnlockImage(ppixOccSynth);
	pimgAlphaMask->UnlockImage(ppixAlphaMask);
	return S_OK;
}
