/**
 *  texsyn.h
 *  Jon Rowlett
 *  11/11/99
 *
 *  Implements much of the Efros-Leung texture synthesis techniques
 *
 **/
#ifndef __TEXSYN_H__
#define __TEXSYN_H__

#include "imgproc.h"

/** 
 *  Region Functions
 *
 *  Functions / classes to "grow" the texture region as described by Efros / Leung
 **/
class CPixelTestIterator
{
public:
	CPixelTestIterator(IImageFactory* pimgf, IImage* pimg);

	virtual ~CPixelTestIterator();

	CPixelTestIterator& operator++();

	void GetTestPixel(int& x, int& y);
	operator bool() const;
	void GetNewMask(IImage** ppimgMask);
	void begin();
	int GetPixelsTested();

protected:
	IImage* m_pimgMask;
	IImageFactory* m_pimgf;
	IImage* m_pimgNewMask;
	int m_nX, m_nY;
	DWORD m_dwCX, m_dwCY;
	BYTE* m_pbLoc;
	BYTE* m_pbNewLoc;
	void* m_pPix;
	void* m_pNewPix;
	int m_nPixelsTested;
};

/**
 *  Window Functions
 *  These functions / classes are used for iterating through texture windows as defined by Efros-Leung
 **/

/**
 *  Base class for all texture window iterators
 *  Motivation is to try several methods to iterate through texture windows
 *  this base provides the framework
 **/
class CTexWindowIterator
{
public:
	CTexWindowIterator(IImage* pimg, int w);
	virtual ~CTexWindowIterator();

	// resets the iterator
	virtual void begin() = 0; 

	// tells when the iterator has finished and the coords are no longer valid
	virtual bool end() = 0;

	// finds next texture window
	virtual CTexWindowIterator& inc() = 0;

	// retrieves the coords of the texture window
	void getCoords(int& x, int& y);
	
protected:
	DWORD m_dwCX, m_dwCY;
	int m_nSize;
	int m_nX;
	int m_nY;

};

/**
 *  Attempt 1 - a rectangle iterator  
 *
 *  This class iterates over a rectangular region
 **/
class CRectTexWindowIterator : public CTexWindowIterator
{
public:
	CRectTexWindowIterator(IImage* pimg, int w, int x, int y, int cx, int cy);
#ifdef WIN32
	CRectTexWindowIterator(IImage* pimg, int w, const RECT& r);
#endif

	// resets the iterator
	virtual void begin(); 

	// tells when the iterator has finished and the coords are no longer valid
	virtual bool end();

	// finds next texture window
	virtual CTexWindowIterator& inc();

protected:
	int m_nX0, m_nY0, m_nX1, m_nY1;
};

#endif /* __TEXSYN_H__*/