#include "imgproc.h"
#include <fstream>
#include <string>
#include <list>
#include <assert.h>
using namespace std;

typedef double vec3[3];

struct triangle
{
	vec3 v[3];
	double getMaxX();
	double getMaxY();
};

double triangle::getMaxX()
{
	return max(max(v[0][0], v[1][0]), v[2][0]);
}

double triangle::getMaxY()
{
	return max(max(v[0][1], v[1][1]), v[2][1]);
}

istream& operator >>(istream& istm, vec3& v)
{
	istm >> v[0] >> v[1] >> v[2];
	v[2] *= 1.0 / 255.0;
	return istm;
}

istream& operator >>(istream& istm, triangle& tri)
{
	string s;
	istm >> s;
	if(!istm)
		return istm;
	if(s != "t")
		istm.setstate(ios::failbit);
	return istm >> tri.v[0] >> tri.v[1] >> tri.v[2];
}

void swapvec3(vec3& v1, vec3& v2)
{
	vec3 tmp;
	memcpy(tmp, v1, sizeof(vec3));
	memcpy(v1, v2, sizeof(vec3));
	memcpy(v2, tmp, sizeof(vec3));
}

void SortTriangle(triangle& tri)
{
	for(int i = 0; i < 2; i++)
		for(int j = i + 1; j < 3; j++)
			if(tri.v[i][1] > tri.v[j][1])
				swapvec3(tri.v[i], tri.v[j]);
}

void memsetgrad(double* pDest, double val1, double dval, int sz)
{
	*pDest = val1;
	for(int j=0; j < sz-1; j++, pDest++)
	{
		pDest[1] = (*pDest) + dval;
	}
}

void imgsetgrad(DWORD dwCX, DWORD dwCY, double* pDest, int y, int x1, int x2, double val1, double val2)
{
	if(y < 0 || y > dwCY - 1)
		return;
	if(x1 > x2)
	{
		int tmp = x1;
		x1 = x2;
		x2 = tmp;
	}
	pDest += y * dwCX;
	double dval = (val2 - val1) / (double)(x2 - x1);
	if(x1 < 0)
	{
		val1 += dval * -x1;
		x1 = 0;
	}
	if(x2 >= dwCX)
	{
		x2 = dwCX-1;
	}
	pDest += x1;
	memsetgrad(pDest, val1, dval, x2 - x1 + 1);
}

class CLine
{
public:
	CLine(int xa, int ya, int xb, int yb)
		: dx(xb - xa), dy(yb - ya), p(0), x(xa), y(ya)
	{
	}

	int getNextX()
	{
		doLine(dx, dy, x, y, p);
		return x;
	}

	int x, y, dx, dy;

protected:
	int p;

	static void doLine(int dx, int dy, int& x, int &y, int &p)
	{
		int i;
		if(abs(dy) > abs(dx))
		{
			if(dx < 0)
			{
				y++;
				p+=dx;
				if(abs(p + dy) < abs(p))
				{
					x--;
					p+=dy;
				}
			}
			else
			{
				y++;
				p+=dx;
				if(abs(p - dy) < abs(p))
				{
					x++;
					p-=dy;
				}
			}
		}
		else
		{
			if(dx < 0)
			{
				for(i=0; i < dx; i++)
				{
					x--;
					p+=dy;
					if(abs(p + dx) < abs(p))
					{
						y++;
						p+=dx;
						return;
					}
				}
			}
			else
			{
				for(i =0; i < dx; i++)
				{
					x++;
					p-=dy;
					if(abs(p + dx) < abs(p))
					{
						y++;
						p+=dx;
						return;
					}
				}
			}
		}
	}
};

void RenderTriangle(triangle& tri, IImage* pimg)
{
	DWORD dwCX, dwCY;
	pimg->GetSize(&dwCX, &dwCY);
	SortTriangle(tri);
	void* ppix;
	pimg->LockImage(&ppix);
	int y;
	double dval1 = (tri.v[1][2] - tri.v[0][2]) / (tri.v[1][1] - tri.v[0][1]);
	double val1 = tri.v[0][2];
	double val2 = val1;
	double val3 = tri.v[1][2];
	double dval2 = (tri.v[2][2] - tri.v[0][2]) / (tri.v[2][1] - tri.v[0][1]);
	double dval3 = (tri.v[2][2] - tri.v[1][2]) / (tri.v[2][1] - tri.v[1][1]);
	CLine l1(tri.v[0][0], tri.v[0][1], tri.v[1][0], tri.v[1][1]);
	CLine l2(tri.v[0][0], tri.v[0][1], tri.v[2][0], tri.v[2][1]);
	CLine l3(tri.v[1][0], tri.v[1][1], tri.v[2][0], tri.v[2][1]);
	for(y = tri.v[0][1]; y < tri.v[1][1]; y++)
	{
		imgsetgrad(dwCX, dwCY, (double*)ppix, y, l1.x, l2.x, val1, val2);
		val1 += dval1;
		val2 += dval2;
		l1.getNextX();
		l2.getNextX();
	}
	for(; y <= tri.v[2][1]; y++)
	{
		imgsetgrad(dwCX, dwCY, (double*)ppix, y, l2.x, l3.x, val2, val3);
		val2 += dval2;
		val3 += dval3;
		l2.getNextX();
		l3.getNextX();
	}
	pimg->UnlockImage(ppix);
}


HRESULT __stdcall LoadTIN(LPCTSTR lpszFilename, IImageFactory* pimgf, IImage** ppimgOut)
{
	ifstream ifs(lpszFilename);
	list<triangle> lsttri;
	triangle tri;
	PixelFormat pf;
	pf.bChannels = 1;
	pf.dwSize = sizeof(PixelFormat);
	pf.dwTypeFlags = PF_LONG | PF_FLOAT;
	double cx = 0.0, cy = 0.0;
	while(!ifs.eof())
	{
		ifs >> tri;
		if(!ifs)
			break;
		double x = tri.getMaxX();
		if(x > cx)
			cx = x;
		double y = tri.getMaxY();
		if(y > cy)
			cy = y;
		lsttri.push_back(tri);
	}
	assert(cx > 0.0);
	assert(cy > 0.0);
	HRESULT hr = pimgf->CreateImage((DWORD)(cx + 0.5), (DWORD)(cy + 0.5), &pf, ppimgOut);
	if(FAILED(hr))
		return hr;
	list<triangle>::iterator it = lsttri.begin();
	for(; it != lsttri.end(); ++it)
	{
		RenderTriangle(*it, *ppimgOut);
	}
	return S_OK;
}
