#include "trans.h"
#include <memory.h>
#include <assert.h>

#define IDX(x,y) ((x) * 3 + (y))
#define PTX(x)   ((x) << 1)
#define PTY(x)   (((x) << 1) + 1)
#define X 0
#define Y 1
#define W 2

#ifndef max
#define max(x,y) ((x) > (y) ? (x) : (y))
#endif

#ifndef min
#define min(x,y) ((x) > (y) ? (x) : (y))
#endif

void CreateWarpMatrix(long pts[8], Matrix3x3 mat)
{
	double dpt[8] = {pts[0], pts[1], pts[2], pts[3], pts[4], pts[5], pts[6], pts[7]};
	CreateWarpMatrix(dpt, mat);
}

void CreateWarpMatrix(double pts[8], Matrix3x3 mat)
{
	double pt0[3] = {pts[PTX(0)], pts[PTY(0)], 1.0};
	double pt1[3] = {pts[PTX(1)], pts[PTY(1)], 1.0};
	double pt2[3] = {pts[PTX(2)], pts[PTY(2)], 1.0};
	double pt3[3] = {pts[PTX(3)], pts[PTY(3)], 1.0};

	double a = pt2[Y] - pt3[Y];
	double b = pt1[Y] - pt2[Y];
	double c = pt0[Y] - pt2[Y];
	double d = pt1[X] * a - pt2[X] * (pt1[Y] - pt3[Y]) + pt3[X] * b;

	double w1 = (pt0[X] * a - (pt2[X] * (pt0[Y] - pt3[Y]) - pt3[X] * c)) / d;
	double w3 = (pt0[X] * b - pt1[X] * c + pt2[X] * (pt0[Y] - pt1[Y])) / d;
	double w2 = w1 + w3 - 1;
	pt1[X] *= w1;
	pt1[Y] *= w1;
	pt1[W] = w1;
	pt2[X] *= w2;
	pt2[Y] *= w2;
	pt2[W] = w2;
	pt3[X] *= w3;
	pt3[Y] *= w3;
	pt3[W] = w3;
	mat[IDX(0,0)] = pt3[X] - pt0[X];
	mat[IDX(0,1)] = pt1[X] - pt0[X];
	mat[IDX(0,2)] = pt0[X];
	mat[IDX(1,0)] = pt3[Y] - pt0[Y];
	mat[IDX(1,1)] = pt1[Y] - pt0[Y];
	mat[IDX(1,2)] = pt0[Y];
	mat[IDX(2,0)] = pt3[W] - pt0[W];
	mat[IDX(2,1)] = pt1[W] - pt0[W];
	mat[IDX(2,2)] = pt0[W];
}

void InvWarpMatrix(Matrix3x3 mat)
{
	double a = mat[IDX(1,1)] * mat[IDX(2,2)] - mat[IDX(1,2)] * mat[IDX(2,1)];
	double b = mat[IDX(0,2)] * mat[IDX(2,1)] - mat[IDX(0,1)] * mat[IDX(2,2)];
	double c = mat[IDX(0,1)] * mat[IDX(1,2)] - mat[IDX(0,2)] * mat[IDX(1,1)];
	double d = mat[IDX(1,2)] * mat[IDX(2,0)] - mat[IDX(1,0)] * mat[IDX(2,2)];
	double e = mat[IDX(0,0)] * mat[IDX(2,2)] - mat[IDX(0,2)] * mat[IDX(2,0)];
	double f = mat[IDX(0,2)] * mat[IDX(1,0)] - mat[IDX(0,0)] * mat[IDX(1,2)];
	double g = mat[IDX(1,0)] * mat[IDX(2,1)] - mat[IDX(1,1)] * mat[IDX(2,0)];
	double h = mat[IDX(0,1)] * mat[IDX(2,0)] - mat[IDX(0,0)] * mat[IDX(2,1)];
	double i = mat[IDX(0,0)] * mat[IDX(1,1)] - mat[IDX(0,1)] * mat[IDX(1,0)];
	double det = mat[IDX(0,0)] * a + mat[IDX(0,1)] * d + mat[IDX(0,2)] * g;
	mat[IDX(0,0)] = a / det;
	mat[IDX(0,1)] = b / det;
	mat[IDX(0,2)] = c / det;
	mat[IDX(1,0)] = d / det;
	mat[IDX(1,1)] = e / det;
	mat[IDX(1,2)] = f / det;
	mat[IDX(2,0)] = g / det;
	mat[IDX(2,1)] = h / det;
	mat[IDX(2,2)] = i / det;
}

void MulWarpMatrix(Matrix3x3 mat1, Matrix3x3 mat2, Matrix3x3 matRes)
{
	matRes[IDX(0,0)] = mat1[IDX(0,0)] * mat2[IDX(0,0)]
		+ mat1[IDX(0,1)] * mat2[IDX(1,0)] + mat1[IDX(0,2)] * mat2[IDX(2,0)];
	matRes[IDX(0,1)] = mat1[IDX(0,0)] * mat2[IDX(0,1)]
		+ mat1[IDX(0,1)] * mat2[IDX(1,1)] + mat1[IDX(0,2)] * mat2[IDX(2,1)];
	matRes[IDX(0,2)] = mat1[IDX(0,0)] * mat2[IDX(0,2)]
		+ mat1[IDX(0,1)] * mat2[IDX(1,2)] + mat1[IDX(0,2)] * mat2[IDX(2,2)];
	matRes[IDX(1,0)] = mat1[IDX(1,0)] * mat2[IDX(0,0)]
		+ mat1[IDX(1,1)] * mat2[IDX(1,0)] + mat1[IDX(1,2)] * mat2[IDX(2,0)];
	matRes[IDX(1,1)] = mat1[IDX(1,0)] * mat2[IDX(0,1)]
		+ mat1[IDX(1,1)] * mat2[IDX(1,1)] + mat1[IDX(1,2)] * mat2[IDX(2,1)];
	matRes[IDX(1,2)] = mat1[IDX(1,0)] * mat2[IDX(0,2)]
		+ mat1[IDX(1,1)] * mat2[IDX(1,2)] + mat1[IDX(1,2)] * mat2[IDX(2,2)];
	matRes[IDX(2,0)] = mat1[IDX(2,0)] * mat2[IDX(0,0)]
		+ mat1[IDX(2,1)] * mat2[IDX(1,0)] + mat1[IDX(2,2)] * mat2[IDX(2,0)];
	matRes[IDX(2,1)] = mat1[IDX(2,0)] * mat2[IDX(0,1)]
		+ mat1[IDX(2,1)] * mat2[IDX(1,1)] + mat1[IDX(2,2)] * mat2[IDX(2,1)];
	matRes[IDX(2,2)] = mat1[IDX(2,0)] * mat2[IDX(0,2)]
		+ mat1[IDX(2,1)] * mat2[IDX(1,2)] + mat1[IDX(2,2)] * mat2[IDX(2,2)];
}

void CreateWarpMapping(double pts0[8], double pts1[8], Matrix3x3 mat)
{
	Matrix3x3 matBase;
	Matrix3x3 matWarp;
	CreateWarpMatrix(pts0, matBase);
	CreateWarpMatrix(pts1, matWarp);
	InvWarpMatrix(matWarp);
	MulWarpMatrix(matBase, matWarp, mat);
}

void TransformPoint(Matrix3x3 mat, double pt[3], double ptOut[3])
{
	ptOut[X] = mat[IDX(0,0)] * pt[X] + mat[IDX(0,1)] * pt[Y] + mat[IDX(0,2)] * pt[W];
	ptOut[Y] = mat[IDX(1,0)] * pt[X] + mat[IDX(1,1)] * pt[Y] + mat[IDX(1,2)] * pt[W];
	ptOut[W] = mat[IDX(2,0)] * pt[X] + mat[IDX(2,1)] * pt[Y] + mat[IDX(2,2)] * pt[W];
}

void TransformPointEx(Matrix3x3 mat, double pt[2], double ptOut[2])
{
	double w;
	ptOut[X] = mat[IDX(0,0)] * pt[X] + mat[IDX(0,1)] * pt[Y] + mat[IDX(0,2)];
	ptOut[Y] = mat[IDX(1,0)] * pt[X] + mat[IDX(1,1)] * pt[Y] + mat[IDX(1,2)];
	w = mat[IDX(2,0)] * pt[X] + mat[IDX(2,1)] * pt[Y] + mat[IDX(2,2)];
	assert(w != 0.0);
	ptOut[X] /= w;
	ptOut[Y] /= w;
}

void InitIdentity(Matrix3x3 mat)
{
	memset(mat, 0, sizeof(double) * 9);
	mat[0] = 1.0;
	mat[4] = 1.0;
	mat[8] = 1.0;

}

void AutoCorrelate2D(double* imgSrc, double* imgDest, unsigned long cx, unsigned long cy)
{
	unsigned long x, y;
	unsigned long w, h;
	double* pSrc1;
	double* pSrc2;
	double* pDest = imgDest;
	pSrc1 = imgSrc;
	memset(pDest, 0, cx * cy * sizeof(double) * 4);
	for(y = 0, h = cy; y < cy; y++, pDest += 4, h--, pSrc1 += 4)
		for(x = 0, w = cx; x < cx; x++, pDest += 4, w--, pSrc1 += 4)
		{
			pSrc2 = imgSrc;
			double* pSrc = pSrc1;
			for(unsigned long j = 0; j < h; j++, pSrc += ((cx - x) << 2), pSrc2 += ((cx - x) << 2))
				for(unsigned long i = 0; i < w; i++, pSrc += 4, pSrc2 += 4)
				{
					pDest[0] += pSrc1[0] * pSrc2[0];
					pDest[1] += pSrc1[1] * pSrc2[1];
					pDest[2] += pSrc1[2] * pSrc2[2];
					pDest[3] += pSrc1[3] * pSrc2[3];
				}
		}
}

/**
 * Convolve2D
 * produces output same size as first image
 * image 2 is a filter applied to image 1
 **/
void Convolve2D(double* pimgSrc1, unsigned long cx, unsigned long cy,
				double* pimgSrc2, unsigned long cx2, unsigned long cy2,
				double* pimgDest, int nClrComp)
{
	unsigned long x, y;
	unsigned long w, h;
	memset(pimgDest, 0, sizeof(double) * cx * cy * nClrComp);
	double* pDest = pimgDest;
	for(y = 0; y < cy; y++, pDest += nClrComp)
		for(x = 0; x < cx; x++, pDest += nClrComp)
		{
			int x1st = max(0, x - cx2);
			int y1st = max(0, y - cy2);
			int x1end = min(cx, x + cx2);
			int y1end = min(cy, y + cy2);
			int x2st = (x + cx2 > cx ? cx - x : cx2) - 1;
			int y2st = (y + cy2 > cy ? cy - y : cy2) - 1;
			int x2end = max(x - cx, 0);
			int y2end = max(y - cy, 0);
			for(int j = 0; j < y1end - y1st; j++)
				for(int i = 0; i < x1end - x1st; i++)
				{
					double* pSrc1 = pimgSrc1 + (cx * (j + y1st) + i + x1st) * 4;
					double* pSrc2 = pimgSrc2 + (cy * (y2st - j) + x2end - i) * 4;
					pDest[0] += pSrc2[0] * pSrc1[0];
					pDest[1] += pSrc2[1] * pSrc1[1];
					pDest[2] += pSrc2[2] * pSrc1[2];
					pDest[3] += pSrc2[3] * pSrc1[3];
				}
		}
}

double DotP(Vector3 v0, Vector3 v1)
{
	return v0[0] * v1[0] + v0[1] * v1[1] + v0[2] * v1[2];
}

void CreateTransMatrix(double x, double y, Matrix3x3 mat)
{
	InitIdentity(mat);
	mat[2] = x;
	mat[5] = y;
}

void CreateScaleMatrix(double sx, double sy, Matrix3x3 mat)
{
	InitIdentity(mat);
	mat[0] = sx;
	mat[4] = sy;
}

bool TexPlaneHitTest(Matrix3x3 mat, double x, double y)
{
	InvWarpMatrix(mat);
	double pt[2] = {x, y};
	double ptout[2];
	TransformPointEx(mat, pt, ptout);
	return (ptout[0] >= 0.0 && ptout[0] < 1.0) &&
		(ptout[1] >= 0.0 && ptout[1] < 1.0);
}

bool TexPlaneHitTestEx(double pts[8], double x, double y)
{
	Matrix3x3 mat;
	CreateWarpMatrix(pts, mat);
	return TexPlaneHitTest(mat, x, y);
}

bool TexPlaneHitTest(Matrix3x3 mat, double x, double y, int nxdiv, int nydiv, int& nx, int& ny)
{
	InvWarpMatrix(mat);
	double pt[2] = {x, y};
	double ptout[2];
	TransformPointEx(mat, pt, ptout);
	if((ptout[0] >= 0.0 && ptout[0] < 1.0) &&
		(ptout[1] >= 0.0 && ptout[1] < 1.0))
	{
		nx = (int)((double)nxdiv * ptout[0]);
		ny = (int)((double)nydiv * ptout[1]);
		return true;
	}
	else
		return false;
}

bool TexPlaneHitTestEx(double pts[8], double x, double y, int nxdiv, int nydiv, int& nx, int& ny)
{
	Matrix3x3 mat;
	CreateWarpMatrix(pts, mat);
	return TexPlaneHitTest(mat, x, y, nxdiv, nydiv, nx, ny);
}

void ManipTexPlane(Matrix3x3 mat, double dx1, double dy1, double dx2, double dy2, double pts[8])
{
	dx2 += 1.0;
	dy2 += 1.0;
	double ptsrc[8] = {dx1, dy1, dx1, dy2, dx2, dy2, dx2, dy1};
	TransformPointEx(mat, ptsrc + 0, pts + 0);
	TransformPointEx(mat, ptsrc + 2, pts + 2);
	TransformPointEx(mat, ptsrc + 4, pts + 4);
	TransformPointEx(mat, ptsrc + 6, pts + 6);
}

void ManipTexPlane(double pts[8], double dx1, double dy1, double dx2, double dy2)
{
	Matrix3x3 mat;
	CreateWarpMatrix(pts, mat);
	ManipTexPlane(mat, dx1, dy1, dx2, dy2, pts);
}

void ManipTexPlane(Matrix3x3 mat, int nxdiv, int nydiv, int dx1, int dy1, int dx2, int dy2, double pts[8])
{
	ManipTexPlane(mat, (double)dx1 / (double)nxdiv, (double)dy1 / (double)nydiv, 
		(double)dx2 / (double)nxdiv, (double)dy2 / (double)nydiv, pts);
}

void ManipTexPlane(double pts[8], int nxdiv, int nydiv, int dx1, int dy1, int dx2, int dy2)
{
	Matrix3x3 mat;
	CreateWarpMatrix(pts, mat);
	ManipTexPlane(mat, nxdiv, nydiv, dx1, dy1, dx2, dy2, pts);
}