/**
 *  trans.h - Transformation Library
 *	Author: Jon Rowlett (rowlett@andrew)
 *  Created: 8/5/99
 **/

#ifndef __TRANS_H__
#define __TRANS_H__

typedef double Matrix3x3[9];
typedef double Vector3[3];

void CreateWarpMatrix(long pts[8], Matrix3x3 mat);
void CreateWarpMatrix(double pts[8], Matrix3x3 mat);
void InvWarpMatrix(Matrix3x3 mat);
void MulWarpMatrix(Matrix3x3 mat1, Matrix3x3 mat2, Matrix3x3 matRes);
void TransformPoint(Matrix3x3 mat, double pt[3], double ptOut[3]);
void TransformPointEx(Matrix3x3 mat, double pt[2], double ptOut[2]);
void InitIdentity(Matrix3x3 mat);

double DotP(Vector3 v0, Vector3 v1);

void CreateWarpMapping(double pts0[8], double pts1[8], Matrix3x3 mat);

void AutoCorrelate2D(double* imgSrc, double* imgDest, unsigned long cx, unsigned long cy);

/**
 * Convolve2D
 * produces output same size as first image
 * image 2 is a filter applied to image 1
 **/
void Convolve2D(double* pimgSrc1, unsigned long cx, unsigned long cy,
				double* pimgSrc2, unsigned long cx2, unsigned long cy2,
				double* pimgDest, int nClrComp);

/**
 *	12/01/99
 *  Homography functions used by the Texture Source Control
 *
 **/

//More Matrix Functions
void CreateTransMatrix(double x, double y, Matrix3x3 mat);
void CreateScaleMatrix(double sx, double sy, Matrix3x3 mat);

bool TexPlaneHitTest(Matrix3x3 mat, double x, double y);
bool TexPlaneHitTestEx(double pts[8], double x, double y);
bool TexPlaneHitTest(Matrix3x3 mat, double x, double y, int nxdiv, int nydiv, int& nx, int& ny);
bool TexPlaneHitTestEx(double pts[8], double x, double y, int nxdiv, int nydiv, int& nx, int& ny);

void ManipTexPlane(Matrix3x3 mat, double dx1, double dy1, double dx2, double dy2, double pts[8]);
void ManipTexPlane(double pts[8], double dx1, double dy1, double dx2, double dy2);
void ManipTexPlane(Matrix3x3 mat, int nxdiv, int nydiv, int dx1, int dy1, int dx2, int dy2, double pts[8]);
void ManipTexPlane(double pts[8], int nxdiv, int nydiv, int dx1, int dy1, int dx2, int dy2);

#endif /* __TRANS_H__ */